      .486
      .model flat, stdcall
      option casemap :none

      include \masm32\include\windows.inc
      include \masm32\include\user32.inc
      include \masm32\include\kernel32.inc
      include \masm32\include\masm32.inc
      include \masm32\macros\macros.asm
      includelib \masm32\lib\masm32.lib
      includelib \masm32\lib\user32.lib
      includelib \masm32\lib\kernel32.lib

      ; Define a data type that includes the first
      ; 16 bytes of the PCI Configuration Data and
      ; the PCI hardware characteristics byte.
      PCICFG STRUCT
         vendorID       dw ?
         deviceID       dw ?
         command_reg    dw ?
         status_reg     dw ?
         revisionID     db ?
         progIF         db ?
         subclass       db ?
         classcode      db ?
         cacheline_size db ?
         latency        db ?
         header_type    db ?
         BIST           db ?
         hardwareChar   db ?
      PCICFG ENDS

.data
      ; This will be the name of the DOS program.
      comFileName db "PCICFG.COM",0

      hFile       dd 0
      bRead       dd 0
      buffer      db 30 dup(0)
.code
start:
      ; Create the DOS program and run it.
      ; No forward references with INVOKE.
      mov   eax,OFFSET comStart
      mov   ecx,OFFSET comEOF - OFFSET comStart
      invoke write_disk_file,ADDR comFileName,eax,ecx
      test  eax,eax
      jnz   @F
      print SADD(13,10,"write_disk_file (PCICFG.COM) failed")
   @@:
      invoke shell,ADDR comFileName
      REPEAT 4
         ; Make up to 4 attempts to open the data file,
         ; waiting 200ms between attempts.
         mov   eax,OFFSET dataFileName
         invoke CreateFile,eax,GENERIC_READ,
            FILE_SHARE_READ,NULL,OPEN_EXISTING,NULL,NULL
         cmp   eax,INVALID_HANDLE_VALUE
         jne   @F
         ; As a debug aid, this will display for each failure.
         print SADD(13,10,"CreateFile (open PCICFG.DAT) failed")
         invoke Sleep,200
      ENDM
      print SADD(13,10,"Could not open PCICFG.DAT")
   @@:
      ; Read parameters into structure and close the file.
      mov   hFile,eax
      mov   eax,OFFSET pcicfg
      mov   ecx,SIZEOF pcicfg
      invoke ReadFile,hFile,eax,ecx,ADDR bRead,NULL
      invoke CloseHandle,hFile

      ; Display selected values.
      
      ; The vendor ID identifies the manufacturer and it
      ; is assigned by a "central authority".
      ; Intel = 8086h
      ; VIA = 1106h
      ; SIS = 1039h
      ; Acer (ALI) = 1025h
      ; AMD = 1022h
      movzx eax,pcicfg.vendorID
      invoke dw2hex,eax,ADDR buffer
      print SADD(13,10,"vendorID = 0x")
      print ADDR buffer

      ; The device ID identifies the type of device and
      ; it is assigned by the manufacturer.
      movzx eax,pcicfg.deviceID
      invoke dw2hex,eax,ADDR buffer
      print SADD(13,10,"deviceID = 0x")
      print ADDR buffer
      
      ; Should return classcode = 6 ("Bridge device") and
      ; subclass = 0 (Host/PCI bridge).
      
      movzx eax,pcicfg.classcode
      invoke dw2hex,eax,ADDR buffer
      print SADD(13,10,"classcode = 0x")
      print ADDR buffer

      movzx eax,pcicfg.subclass
      invoke dw2hex,eax,ADDR buffer
      print SADD(13,10,"subclass = 0x")
      print ADDR buffer

      ; PCIhardware characteristics byte:
      ; bit0 configuration space access mechanism 1 supported
      ; bit1 configuration space access mechanism 2 supported
      ; bit2-3 reserved
      ; bit4 special cycle generation mechanism 1 supported
      ; bit5 special cycle generation mechanism 2 supported
      ; bit6-7 reserved
      movzx eax,pcicfg.hardwareChar
      invoke dw2hex,eax,ADDR buffer
      print SADD(13,10,"hardware characteristics = 0x")
      print ADDR buffer

      print SADD(13,10,"Press enter to exit...",13,10)
      invoke StdIn,addr buffer,1

      ; Delete the DOS program and the data file.
      invoke DeleteFile,ADDR comFileName
      test   eax,eax
      jnz   @F
      print SADD(13,10,"DeleteFile (PCICFG.COM) failed")
   @@:
      mov   eax,OFFSET dataFileName
      invoke DeleteFile,eax
      test   eax,eax
      jnz   @F
      print SADD(13,10,"DeleteFile (PCICFG.DAT) failed")
   @@:
      exit   

; End the 32-bit segment created by the .MODEL directive.
_TEXT ENDS

; Start a 16-bit segment.
_TEXT16 SEGMENT USE16

comStart:
      ; Start of the DOS program that will actually call the
      ; BIOS function. The program will terminate on the first
      ; error.

      ; Jump around data.
      jmp   @F

      pcicfg         PCICFG <>
      dataFileName   db "PCICFG.DAT",0

      ; Windows will load the program into memory starting at
      ; offset 100h. Calculate the run-time offset addresses.
      OFFSET_PCICFG EQU 100h + OFFSET pcicfg - OFFSET comStart
      OFFSET_DATAFILENAME EQU 100h + OFFSET dataFileName - OFFSET comStart
   @@:
      ; *** Removing this for XP test, did not correct problem ***
      ; Make sure the PCI BIOS (v2.0c+) is installed
      ; and save the hardware characteristics byte.
      mov   ax,0b101h
      mov   edi,0
      int   1ah
      test  ah,ah
      jnz   comEnd      ; Not installed.
      ; ************
      ; Had to use EBX to avoid "error A2155 "cannot use
      ; 16-bit register with a 32-bit address"
      ; But USE16 is supposed to specify a 16-bit
      ; operand and address size?
      ; ************
      mov   ebx,OFFSET_PCICFG+16
      mov   [ebx],al

      ; Get the first 16 bytes of the PCI Configuration
      ; Data directly from the PCI Configuration
      ; Mechanism 1 I/O ports.

      ADDRESS_PORT   EQU 0cf8h
      DATA_PORT      EQU 0cfch

      ; Address bitfields:
      ; 1-0    reserved (00)
      ; 7-2    configuration register number
      ; 10-8   function
      ; 15-11  device number
      ; 23-16  bus number
      ; 30-24  reserved (0)
      ; 31     enable configuration space mapping
      xor   ebx,ebx
   @@:
      ; Preserve the original address.
      mov   dx,ADDRESS_PORT
      in    eax,dx
      push  eax
      ; Set up and output the new address.
      mov   eax,ebx
      shr   eax,2
      shl   eax,2
      or    eax,80000000h
      out   dx,eax
      ; Input the register value.
      mov   dx,DATA_PORT
      in    eax,dx
      ; Copy to structure.
      mov   [OFFSET_PCICFG+ebx],eax
      ; Restore the original address.
      pop   eax
      mov   dx,ADDRESS_PORT
      out   dx,eax
      add   ebx,4
      cmp   ebx,16
      jb    @B

      ; Create the data file and write the structure to it.
      ATTR_NORMAL EQU 0
      mov   dx,OFFSET_DATAFILENAME
      mov   cx,ATTR_NORMAL
      mov   ah,3ch
      int   21h
      jc   comEnd

      mov   bx,ax       ; handle
      mov   cx,SIZEOF PCICFG
      mov   dx,OFFSET_PCICFG
      mov   ah,40h
      int   21h

      ; Handle will be closed and file buffers flushed
      ; on terminate.
   comEnd:
      mov   ax,4c00h
      int   21h

comEOF:

; This was added in a failed attempt to
; correct the address size problem.
_TEXT16 ENDS

end start
